{********************************* i n f o **********************************}
{*                          } unit Translator; {                            *}
{*                          (c) 2000 Michal Cihar                           *}
{******************************** e m a i l *********************************}
{*                            <cihar@email.cz>                              *}
{***************************** i n t e r n e t ******************************}
{*                          http://cihar.liten.cz                           *}
{********************************** i c q ***********************************}
{*                              UIN: 33471395                               *}
{************************** d e s c r i p t i o n ***************************}
{*    This unit includes component TTranslator. TTranslator can be used for *}
{*  internationalizing your application. It can automatically change many   *}
{*  properties (Caption, Hint, Title, DisplayLabel, Filter, Lines, Items,   *}
{*  CustomColors, Pages and Tabs) by reading it's value from ini file.      *}
{******************************* m a n u a l ********************************}
{*                                                                          *}
{*                                                                          *}
{****************************** l i c e n c e *******************************}
{*    This source is freeware with some small restrictions. For details     *}
{*  look into file LICENCE.TXT, which you should get with this source.      *}
{****************************** h i s t o r y *******************************}
{*  version : date                                                          *}
{*            changes (+ added, - removed, * changed)                       *}
{*  1.0 : 30.8.2000                                                         *}
{*        + first released version                                          *}
{****************************************************************************}
interface

uses Classes,
  Forms,
  MyIniFiles,
  TypInfo,
  SysUtils,
  Windows;

{ Message dialog }

type
  TMsgDlgType = (mtWarning, mtError, mtInformation, mtConfirmation, mtCustom);
  TMsgDlgBtn = (mbYes, mbNo, mbOK, mbCancel, mbAbort, mbRetry, mbIgnore,
    mbAll, mbNoToAll, mbYesToAll, mbHelp);
  TMsgDlgButtons = set of TMsgDlgBtn;
  TButtonCaptions = array[TMsgDlgBtn] of string;
  TCaptions = array[TMsgDlgType] of string;
const
  mbYesNoCancel = [mbYes, mbNo, mbCancel];
  mbOKCancel = [mbOK, mbCancel];
  mbAbortRetryIgnore = [mbAbort, mbRetry, mbIgnore];

  mrNone = 0;
  mrOk = idOk;
  mrCancel = idCancel;
  mrAbort = idAbort;
  mrRetry = idRetry;
  mrIgnore = idIgnore;
  mrYes = idYes;
  mrNo = idNo;
  mrAll = mrNo + 1;
  mrNoToAll = mrAll + 1;
  mrYesToAll = mrNoToAll + 1;

type
  TTranslatorNotifyEvent = procedure(Sender: TObject;
    AComponent: TComponent;
    PropertyName, OldValue: string;
    var NewValue: string) of object;

type
  TTranslator = class(TComponent)
  private
    FIniFile: string;

    FListSeparator: Char;
    FCRChar: Char;
    FLFChar: Char;
    FTabChar: Char;

    FTranslationsSection: string;
    FMessagesSection: string;
    FInfoSection: string;
    FDialogsSection: string;

    FUseEmpty: Boolean;
    FRootComponent: TComponent;
    FAddRootName: Boolean;

    FRefreshInfo: Boolean;
    FRefreshDialogs: Boolean;

    FCaptions: TCaptions;
    FButtonCaptions: TButtonCaptions;
    ButtonWidths: array[TMsgDlgBtn] of Integer;

    FTranslationAuthor: string;
    FTranslationName: string;

    FOnBeforeTranslation: TNotifyEvent;
    FOnAfterTranslation: TNotifyEvent;
    FOnTranslate: TTranslatorNotifyEvent;
    function IfProperty(C: TComponent; Name: string): Boolean;
    function GetComponentProp(C: TComponent; Name: string): string;
    procedure ReadSubStrings(AComponent: TComponent; Path: string; IniFile: TMyIniFile; Section:
      string);
    procedure WriteSubStrings(AComponent: TComponent; Path: string; IniFile: TMyIniFile; Section:
      string);
    function RemoveCRLF(S: string): string;
    function AddCRLF(S: string): string;
    procedure RefreshDialogs;
    procedure RefreshInfo;
    procedure SetIniFile(AValue: string);
    procedure SetInfoSection(AValue: string);
    procedure SetDialogsSection(AValue: string);
    function GetTranslationAuthor: string;
    function GetTranslationName: string;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure CreateIniFile;
    procedure Translate;
    function TranslateUserMessage(const sMsgName, sMessage: string): string;

{ Localisable message dialog }
    function CreateMessageDialog(const Msg: string; DlgType: TMsgDlgType;
      Buttons: TMsgDlgButtons): TForm;

    function MessageDlg(const Msg: string; DlgType: TMsgDlgType;
      Buttons: TMsgDlgButtons; HelpCtx: Longint): Integer;
    function MessageDlgPos(const Msg: string; DlgType: TMsgDlgType;
      Buttons: TMsgDlgButtons; HelpCtx: Longint; X, Y: Integer): Integer;
    function MessageDlgPosHelp(const Msg: string; DlgType: TMsgDlgType;
      Buttons: TMsgDlgButtons; HelpCtx: Longint; X, Y: Integer;
      const HelpFileName: string): Integer;

    procedure ShowMessage(const Msg: string);
    procedure ShowMessageFmt(const Msg: string; Params: array of const);
    procedure ShowMessagePos(const Msg: string; X, Y: Integer);

{ Localisable input dialog }

    function InputBox(const ACaption, APrompt, ADefault: string): string;
    function InputQuery(const ACaption, APrompt: string;
      var Value: string): Boolean;

  published
    property LanguageFile: string read FIniFile write SetIniFile;

    property RootComponent: TComponent read FRootComponent write FRootComponent;
    property AddRootName: Boolean read FAddRootName write FAddRootName default True;
    property UseEmpty: Boolean read FUseEmpty write FUseEmpty default False;

    property OnBeforeTranslation: TNotifyEvent read FOnBeforeTranslation write
    FOnBeforeTranslation;
    property OnAfterTranslation: TNotifyEvent read FOnAfterTranslation write FOnAfterTranslation;
    property OnTranslate: TTranslatorNotifyEvent read FOnTranslate write FOnTranslate;

    property TranslationsSection: string read FTranslationsSection write FTranslationsSection;
    property MessagesSection: string read FMessagesSection write FMessagesSection;
    property InfoSection: string read FInfoSection write SetInfoSection;
    property DialogsSection: string read FDialogsSection write SetDialogsSection;

    property ListSeparator: Char read FListSeparator write FListSeparator default #133;
    property CRChar: Char read FCRChar write FCRChar default #215;
    property LFChar: Char read FLFChar write FLFChar default #247;
    property TabChar: Char read FTabChar write FTabChar default #183;

    property TranslationAuthor: string read GetTranslationAuthor;
    property TranslationName: string read GetTranslationName;

  end;

function TranslateUserMessageFromIni(const FIniFile, FMessagesSection, sMsgName, sMessage: string; const FCRChar, FLFChar, FTabChar: Char): string; overload;
{
Sample ussage:
TranslateUserMessageFromIni('language.ini', 'Messages', '', '', #215, #247, #183);
}

function TranslateUserMessageFromIni(const FIniFile, sMsgName, sMessage: string): string; overload;


procedure Register;

implementation

uses ComCtrls,
  Graphics, {TCanvas}
  ExtCtrls, {TImage}
  StdCtrls; {TLabel, TButton}

{=====================}
const
  DefaultCaptions: TCaptions =
  ('Warning',
    'Error',
    'Information',
    'Confirm',
    '');
  DefaultButtonCaptions: TButtonCaptions =
  ('&Yes',
    '&No',
    'OK',
    'Cancel',
    '&Abort',
    '&Retry',
    '&Ignore',
    '&All',
    'N&o to All',
    'Yes to &All',
    '&Help');

  IconIDs: array[TMsgDlgType] of PChar = (IDI_EXCLAMATION, IDI_HAND,
    IDI_ASTERISK, IDI_QUESTION, nil);
  ButtonNames: array[TMsgDlgBtn] of string = (
    'Yes', 'No', 'OK', 'Cancel', 'Abort', 'Retry', 'Ignore', 'All', 'NoToAll',
    'YesToAll', 'Help');
  ModalResults: array[TMsgDlgBtn] of Integer = (
    mrYes, mrNo, mrOk, mrCancel, mrAbort, mrRetry, mrIgnore, mrAll, mrNoToAll,
    mrYesToAll, 0);




procedure Register;
begin
  RegisterComponents('Michal Cihar', [TTranslator]);
end;
{=====================}

constructor TTranslator.Create(AOwner: TComponent);
var
  DB: TMsgDlgBtn;
begin
  inherited Create(AOwner);
  FMessagesSection := 'Messages';
  FTranslationsSection := 'Translations';
  FDialogsSection := 'Dialogs';
  FInfoSection := 'Info';
  FUseEmpty := False;
  FAddRootName := True;
  FRefreshInfo := True;
  FRefreshDialogs := True;
  FCRChar := #215; {}
  FLFChar := #247; {}
  FTabChar := #183; {}
  FListSeparator := #133; {}
  for DB := mbYes to mbHelp do
    ButtonWidths[DB] := 0;
end;
{=====================}

destructor TTranslator.Destroy;
begin
  inherited Destroy;
end;
{=====================}

procedure TTranslator.RefreshDialogs;
var
  DT: TMsgDlgType;
  DB: TMsgDlgBtn;
begin
  try
    with TMyIniFile.Create(FIniFile) do
    try
      for DT := mtWarning to mtConfirmation do
        FCaptions[DT] := ReadString(FDialogsSection, DefaultCaptions[DT], DefaultCaptions[DT]);
      for DB := mbYes to mbHelp do
        FButtonCaptions[DB] := ReadString(FDialogsSection, DefaultButtonCaptions[DB],
          DefaultButtonCaptions[DB]);
    finally
      Free;
    end;
  except
    FCaptions := DefaultCaptions;
    FButtonCaptions := DefaultButtonCaptions;
  end;
  for DB := mbYes to mbHelp do
    ButtonWidths[DB] := 0;
end;

procedure TTranslator.SetIniFile(AValue: string);
begin
  FRefreshInfo := True;
  FRefreshDialogs := True;
  FIniFile := AValue;
end;

procedure TTranslator.SetInfoSection(AValue: string);
begin
  FRefreshInfo := True;
  FInfoSection := AValue;
end;

procedure TTranslator.SetDialogsSection(AValue: string);
begin
  FRefreshDialogs := True;
  FDialogsSection := AValue;
end;

function TTranslator.GetTranslationAuthor: string;
begin
  if FRefreshInfo then RefreshInfo;
  Result := FTranslationAuthor;
end;

procedure TTranslator.RefreshInfo;
begin
  try
    with TMyIniFile.Create(FIniFile) do
    try
      FTranslationAuthor := ReadString(FTranslationName, 'Author', '');
      FTranslationName := ReadString(FTranslationName, 'Language', '');
    finally
      Free;
    end;
  except
    FTranslationAuthor := '';
    FTranslationName := '';
  end;
end;

function TTranslator.GetTranslationName: string;
begin
  if FRefreshInfo then RefreshInfo;
  Result := FTranslationName;
end;


procedure TTranslator.Translate;
var
  MyIniFile: TMyIniFile;
begin
  try
    MyIniFile := TMyIniFile.Create(FIniFile);
    try
      if (Assigned(FOnBeforeTranslation)) then FOnBeforeTranslation(Self);
      if Assigned(FRootComponent) then
        if FAddRootName then
          ReadSubStrings(FRootComponent, FRootComponent.Name, MyIniFile, FTranslationsSection)
        else
          ReadSubStrings(FRootComponent, '', MyIniFile, FTranslationsSection)
      else
        if FAddRootName then
          ReadSubStrings(Application, 'Application', MyIniFile, FTranslationsSection)
        else
          ReadSubStrings(Application, '', MyIniFile, FTranslationsSection);
    finally
      MyIniFile.Free;
      if (Assigned(FOnAfterTranslation)) then FOnAfterTranslation(Self);
    end;
  except
  end;
end;
{=====================}


function TTranslator.TranslateUserMessage(const sMsgName, sMessage: string): string;
begin
  try
    with TMyIniFile.Create(FIniFile) do
    try
      Result := AddCRLF(ReadString(FMessagesSection, RemoveCRLF(sMsgName), '---- ---- ----'));
      if Result = '---- ---- ----' then
        Result := AddCRLF(ReadString(FMessagesSection, RemoveCRLF(sMessage), RemoveCRLF(sMessage)));
    finally
      Free;
    end;
  except
    Result := sMessage;
  end;
end;
{=====================}

function TTranslator.IfProperty(C: TComponent; Name: string): Boolean;
begin
  Result := (C.Name <> '') and (GetPropInfo(C.ClassInfo, Name) <> nil);
end;

function TTranslator.GetComponentProp(C: TComponent; Name: string): string;
var
  P: PPropInfo;
  propValue: TObject;
  J: Integer;
begin
  Result := '';
  P := GetPropInfo(C.ClassInfo, Name);
  if Assigned(P) then
  begin
    if (P^.PropType^)^.Kind in [tkString, tkWChar, tkLString, tkWString] then
      Result := GetStrProp(C, P)
    else
      if (P^.PropType^)^.Kind in [tkClass] then
      begin
        propValue := TObject(GetOrdProp(C, P));
        if (propValue is TStrings) then
          Result := TStrings(propValue).CommaText
        else
          if (propValue is TTreeNodes) then
            for J := 0 to TTreeNodes(propValue).Count - 1 do
              Result := Result + TTreeNodes(propValue).Item[J].Text + FListSeparator;
      end;
  end;
end;

function TTranslator.RemoveCRLF(S: string): string;
var
  I: Integer;
begin
  for I := 1 to Length(S) do
    if S[I] = #10 then
      S[I] := FLFChar
    else
      if S[I] = #13 then
        S[I] := FCRChar
      else
        if S[I] = #9 then S[I] := FTabChar;
  Result := S;
end;

function TTranslator.AddCRLF(S: string): string;
var
  I: Integer;
begin
  for I := 1 to Length(S) do
    if S[I] = FLFChar then
      S[I] := #10
    else
      if S[I] = FCRChar then
        S[I] := #13
      else
        if S[I] = FTabChar then S[I] := #9;
  Result := S;
end;


procedure TTranslator.ReadSubStrings(AComponent: TComponent; Path: string; IniFile: TMyIniFile;
  Section: string);
var
  S: string;
  I: Integer;

  procedure DoProperty(Which: string);
  var
    P: PPropInfo;
    propValue: TObject;
    J: Integer;
  begin
    if (IfProperty(AComponent, Which)) then
    begin
      S := IniFile.ReadString(Section, Path + Which, '');
      S := AddCRLF(S);
      if (FUseEmpty) or ((Trim(S) <> '') and (Trim(S) <> '-')) then
      begin
        if Assigned(FOnTranslate) then
          FOnTranslate(Self, AComponent, Which, GetComponentProp(AComponent, Which), S);
        P := GetPropInfo(AComponent.ClassInfo, Which);
        if Assigned(P) then
        begin
          if (P^.PropType^)^.Kind in [tkString, tkWChar, tkLString, tkWString] then
            SetStrProp(AComponent, P, S)
          else
            if (P^.PropType^)^.Kind in [tkClass] then
            begin
              propValue := TObject(GetOrdProp(AComponent, P));
              if (propValue is TStrings) then
              begin
                if Trim(S)[1] <> '"' then S := '"' + S + '"'; //this is unusual but works...
                TStrings(propValue).CommaText := S;
              end
              else
                if (propValue is TTreeNodes) then
                begin
                  J := 0;
                  while (Pos(FListSeparator, S) > 0) and (J < TTreeNodes(propValue).Count) do
                  begin
                    TTreeNodes(propValue).Item[J].Text := Copy(S, 1, Pos(FListSeparator,
                      S) - 1);
                    Delete(S, 1, Pos(FListSeparator, S));
                    Inc(J);
                  end;
                end;
            end;
        end;
      end;
    end;
  end;

begin
  if Trim(Path) <> '' then Path := Path + '.';
  if ((-AComponent.Tag and $80) <> $80) then
  begin
//Text items:
    if ((-AComponent.Tag and $40) = $40) then
      DoProperty('Hint')
    else
    begin
      if ((-AComponent.Tag and $10) <> $10) then DoProperty('Caption');
      if ((-AComponent.Tag and $20) <> $20) then DoProperty('Hint');
      DoProperty('Title');
      DoProperty('DisplayLabel');
      DoProperty('Filter');

//TStrings items:
      DoProperty('Lines');
      DoProperty('Items');
      DoProperty('CustomColors');
      DoProperty('Pages');
      DoProperty('Tabs');
    end;
//      DoProperty('');
  end;
//process subcomponents:
  if ((-AComponent.Tag and $100) <> $100) then
  begin
    for I := 0 to AComponent.ComponentCount - 1 do
    begin
      ReadSubStrings(AComponent.Components[I], Path + AComponent.Components[I].Name,
        IniFile, Section);
    end;
  end;
end;


procedure TTranslator.WriteSubStrings(AComponent: TComponent; Path: string; IniFile:
  TMyIniFile; Section: string);
var
  S: string;
  I: Integer;

  procedure DoProperty(Which: string);
  var
    P: PPropInfo;
    propValue: TObject;
    J: Integer;
  begin
    if (IfProperty(AComponent, Which)) then
    begin
      S := '';
      P := GetPropInfo(AComponent.ClassInfo, Which);
      if Assigned(P) then
      begin
        if (P^.PropType^)^.Kind in [tkString, tkWChar, tkLString, tkWString] then
          S := GetStrProp(AComponent, P)
        else
          if (P^.PropType^)^.Kind in [tkClass] then
          begin
            propValue := TObject(GetOrdProp(AComponent, P));
            if (propValue is TStrings) then
              S := TStrings(propValue).CommaText
            else
              if (propValue is TTreeNodes) then
                for J := 0 to TTreeNodes(propValue).Count - 1 do
                  S := S + TTreeNodes(propValue).Item[J].Text + FListSeparator;
          end;
      end;

      S := RemoveCRLF(S);
      if (FUseEmpty) or ((Trim(S) <> '') and (Trim(S) <> '-')) then
        IniFile.WriteString(Section, Path + Which, S);
    end;
  end;

begin
  if Trim(Path) <> '' then Path := Path + '.';
  if ((-AComponent.Tag and $80) <> $80) then
  begin
//Text items:
    if ((-AComponent.Tag and $40) = $40) then
      DoProperty('Hint')
    else
    begin
      if ((-AComponent.Tag and $10) <> $10) then DoProperty('Caption');
      if ((-AComponent.Tag and $20) <> $20) then DoProperty('Hint');
      DoProperty('Title');
      DoProperty('DisplayLabel');
      DoProperty('Filter');

//TStrings items:
      DoProperty('Lines');
      DoProperty('Items');
      DoProperty('CustomColors');
      DoProperty('Pages');
      DoProperty('Tabs');
    end;
//      DoProperty('');
  end;
//process subcomponents:
  if ((-AComponent.Tag and $100) <> $100) then
  begin
    for I := 0 to AComponent.ComponentCount - 1 do
    begin
      WriteSubStrings(AComponent.Components[I], Path + AComponent.Components[I].Name,
        IniFile, Section);
    end;
  end;
end;

procedure TTranslator.CreateIniFile;
var
  MyIniFile: TMyIniFile;
begin
  try
    MyIniFile := TMyIniFile.Create(FIniFile);
    try
      if Assigned(FRootComponent) then
        if FAddRootName then
          WriteSubStrings(FRootComponent, FRootComponent.Name, MyIniFile, FTranslationsSection)
        else
          WriteSubStrings(FRootComponent, '', MyIniFile, FTranslationsSection)
      else
        if FAddRootName then
          WriteSubStrings(Application, 'Application', MyIniFile, FTranslationsSection)
        else
          WriteSubStrings(Application, '', MyIniFile, FTranslationsSection);
    finally
      MyIniFile.Free;
    end;
  except
  end;
end;
{=========================================================================================================}
{=========================================================================================================}
{=========================================================================================================}
{=========================================================================================================}
{=========================================================================================================}
{=========================================================================================================}
{=========================================================================================================}
{=========================================================================================================}
{=========================================================================================================}
{=========================================================================================================}
{ Message dialog }

function Max(I, J: Integer): Integer;
begin
  if I > J then
    Result := I
  else
    Result := J;
end;

function GetAveCharSize(Canvas: TCanvas): TPoint;
var
  I: Integer;
  Buffer: array[0..51] of Char;
begin
  for I := 0 to 25 do
    Buffer[I] := Chr(I + Ord('A'));
  for I := 0 to 25 do
    Buffer[I + 26] := Chr(I + Ord('a'));
  GetTextExtentPoint(Canvas.Handle, Buffer, 52, TSize(Result));
  Result.X := Result.X div 52;
end;


type
  TMessageForm = class(TForm)
  private
    procedure HelpButtonClick(Sender: TObject);
  public
    constructor CreateNew(AOwner: TComponent); reintroduce;
  end;

constructor TMessageForm.CreateNew(AOwner: TComponent);
var
  NonClientMetrics: TNonClientMetrics;
begin
  inherited CreateNew(AOwner);
  NonClientMetrics.cbSize := SizeOf(NonClientMetrics);
  if SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, @NonClientMetrics, 0) then
    Font.Handle := CreateFontIndirect(NonClientMetrics.lfMessageFont);
end;

procedure TMessageForm.HelpButtonClick(Sender: TObject);
begin
  Application.HelpContext(HelpContext);
end;


{var
{   Captions: array[TMsgDlgType] of PString = (@SMsgDlgWarning, @SMsgDlgError,
      @SMsgDlgInformation, @SMsgDlgConfirm, @SMsgDlgCustom);}
{  Captions: array[TMsgDlgType] of Pointer = (@SMsgDlgWarning, @SMsgDlgError,
    @SMsgDlgInformation, @SMsgDlgConfirm, nil);}
{   IconIDs: array[TMsgDlgType] of PChar = (IDI_EXCLAMATION, IDI_HAND,
      IDI_ASTERISK, IDI_QUESTION, nil);
   ButtonNames: array[TMsgDlgBtn] of string = (
      'Yes', 'No', 'OK', 'Cancel', 'Abort', 'Retry', 'Ignore', 'All', 'NoToAll',
      'YesToAll', 'Help');}
{   ButtonCaptions: array[TMsgDlgBtn] of PString = (
      @SMsgDlgYes, @SMsgDlgNo, @SMsgDlgOK, @SMsgDlgCancel, @SMsgDlgAbort,
      @SMsgDlgRetry, @SMsgDlgIgnore, @SMsgDlgAll, @SMsgDlgNoToAll, @SMsgDlgYesToAll,
      @SMsgDlgHelp);}
{  ButtonCaptions: array[TMsgDlgBtn] of Pointer = (
    @SMsgDlgYes, @SMsgDlgNo, @SMsgDlgOK, @SMsgDlgCancel, @SMsgDlgAbort,
    @SMsgDlgRetry, @SMsgDlgIgnore, @SMsgDlgAll, @SMsgDlgNoToAll, @SMsgDlgYesToAll,
    @SMsgDlgHelp);}
{   ModalResults: array[TMsgDlgBtn] of Integer = (
      mrYes, mrNo, mrOk, mrCancel, mrAbort, mrRetry, mrIgnore, mrAll, mrNoToAll,
      mrYesToAll, 0);}

function TTranslator.CreateMessageDialog(const Msg: string; DlgType: TMsgDlgType;
  Buttons: TMsgDlgButtons): TForm;
const
  mcHorzMargin = 8;
  mcVertMargin = 8;
  mcHorzSpacing = 10;
  mcVertSpacing = 10;
  mcButtonWidth = 50;
  mcButtonHeight = 14;
  mcButtonSpacing = 4;
var
  DialogUnits: TPoint;
  HorzMargin, VertMargin, HorzSpacing, VertSpacing, ButtonWidth,
    ButtonHeight, ButtonSpacing, ButtonCount, ButtonGroupWidth,
    IconTextWidth, IconTextHeight, X, ALeft: Integer;
  B, DefaultButton, CancelButton: TMsgDlgBtn;
  IconID: PChar;
  TextRect: TRect;
begin
  if FRefreshDialogs = True then RefreshDialogs;
  Result := TMessageForm.CreateNew(Application);
  with Result do
  begin
    BiDiMode := Application.BiDiMode;
    BorderStyle := bsDialog;
    Canvas.Font := Font;
    DialogUnits := GetAveCharSize(Canvas);
    HorzMargin := MulDiv(mcHorzMargin, DialogUnits.X, 4);
    VertMargin := MulDiv(mcVertMargin, DialogUnits.Y, 8);
    HorzSpacing := MulDiv(mcHorzSpacing, DialogUnits.X, 4);
    VertSpacing := MulDiv(mcVertSpacing, DialogUnits.Y, 8);
    ButtonWidth := MulDiv(mcButtonWidth, DialogUnits.X, 4);
    for B := Low(TMsgDlgBtn) to High(TMsgDlgBtn) do
    begin
      if B in Buttons then
      begin
        if ButtonWidths[B] = 0 then
        begin
          TextRect := Rect(0, 0, 0, 0);
          Windows.DrawText(Canvas.Handle, PChar(FButtonCaptions[B])
            , -1,
            TextRect, DT_CALCRECT or DT_LEFT or DT_SINGLELINE or
            DrawTextBiDiModeFlagsReadingOnly);
          with TextRect do
            ButtonWidths[B] := Right - Left + 8;
        end;
        if ButtonWidths[B] > ButtonWidth then
          ButtonWidth := ButtonWidths[B];
      end;
    end;
    ButtonHeight := MulDiv(mcButtonHeight, DialogUnits.Y, 8);
    ButtonSpacing := MulDiv(mcButtonSpacing, DialogUnits.X, 4);
    SetRect(TextRect, 0, 0, Screen.Width div 2, 0);
    DrawText(Canvas.Handle, PChar(Msg), Length(Msg), TextRect,
      DT_EXPANDTABS or DT_CALCRECT or DT_WORDBREAK or
      DrawTextBiDiModeFlagsReadingOnly);
    IconID := IconIDs[DlgType];
    IconTextWidth := TextRect.Right;
    IconTextHeight := TextRect.Bottom;
    if IconID <> nil then
    begin
      Inc(IconTextWidth, 32 + HorzSpacing);
      if IconTextHeight < 32 then IconTextHeight := 32;
    end;
    ButtonCount := 0;
    for B := Low(TMsgDlgBtn) to High(TMsgDlgBtn) do
      if B in Buttons then Inc(ButtonCount);
    ButtonGroupWidth := 0;
    if ButtonCount <> 0 then
      ButtonGroupWidth := ButtonWidth * ButtonCount +
        ButtonSpacing * (ButtonCount - 1);
    ClientWidth := Max(IconTextWidth, ButtonGroupWidth) + HorzMargin * 2;
    ClientHeight := IconTextHeight + ButtonHeight + VertSpacing +
      VertMargin * 2;
    Left := (Screen.Width div 2) - (Width div 2);
    Top := (Screen.Height div 2) - (Height div 2);
    if DlgType <> mtCustom then
      Caption := FCaptions[DlgType]
    else
      Caption := Application.Title;
    if IconID <> nil then
      with TImage.Create(Result) do
      begin
        Name := 'Image';
        Parent := Result;
        Picture.Icon.Handle := LoadIcon(0, IconID);
        SetBounds(HorzMargin, VertMargin, 32, 32);
      end;
    with TLabel.Create(Result) do
    begin
      Name := 'Message';
      Parent := Result;
      WordWrap := True;
      Caption := Msg;
      BoundsRect := TextRect;
      BiDiMode := Result.BiDiMode;
      ALeft := IconTextWidth - TextRect.Right + HorzMargin;
      if UseRightToLeftAlignment then
        ALeft := Result.ClientWidth - ALeft - Width;
      SetBounds(ALeft, VertMargin,
        TextRect.Right, TextRect.Bottom);
    end;
    if mbOK in Buttons then
      DefaultButton := mbOK
    else
      if mbYes in Buttons then
        DefaultButton := mbYes
      else
        DefaultButton := mbRetry;
    if mbCancel in Buttons then
      CancelButton := mbCancel
    else
      if mbNo in Buttons then
        CancelButton := mbNo
      else
        CancelButton := mbOK;
    X := (ClientWidth - ButtonGroupWidth) div 2;
    for B := Low(TMsgDlgBtn) to High(TMsgDlgBtn) do
      if B in Buttons then
        with TButton.Create(Result) do
        begin
          Name := ButtonNames[B];
          Parent := Result;
          Caption := PChar(FButtonCaptions[B]);
          ModalResult := ModalResults[B];
          if B = DefaultButton then default := True;
          if B = CancelButton then Cancel := True;
          SetBounds(X, IconTextHeight + VertMargin + VertSpacing,
            ButtonWidth, ButtonHeight);
          Inc(X, ButtonWidth + ButtonSpacing);
          if B = mbHelp then
            OnClick := TMessageForm(Result).HelpButtonClick;
        end;
  end;
end;

function TTranslator.MessageDlg(const Msg: string; DlgType: TMsgDlgType;
  Buttons: TMsgDlgButtons; HelpCtx: Longint): Integer;
begin
  Result := MessageDlgPosHelp(Msg, DlgType, Buttons, HelpCtx, -1, -1, '');
end;

function TTranslator.MessageDlgPos(const Msg: string; DlgType: TMsgDlgType;
  Buttons: TMsgDlgButtons; HelpCtx: Longint; X, Y: Integer): Integer;
begin
  Result := MessageDlgPosHelp(Msg, DlgType, Buttons, HelpCtx, X, Y, '');
end;

function TTranslator.MessageDlgPosHelp(const Msg: string; DlgType: TMsgDlgType;
  Buttons: TMsgDlgButtons; HelpCtx: Longint; X, Y: Integer;
  const HelpFileName: string): Integer;
begin
  with CreateMessageDialog(Msg, DlgType, Buttons) do
  try
    HelpContext := HelpCtx;
    HelpFile := HelpFileName;
    if X >= 0 then Left := X;
    if Y >= 0 then Top := Y;
    if (Y < 0) and (X < 0) then Position := poScreenCenter;
    Result := ShowModal;
  finally
    Free;
  end;
end;

procedure TTranslator.ShowMessage(const Msg: string);
begin
  ShowMessagePos(Msg, -1, -1);
end;

procedure TTranslator.ShowMessageFmt(const Msg: string; Params: array of const);
begin
  ShowMessage(Format(Msg, Params));
end;

procedure TTranslator.ShowMessagePos(const Msg: string; X, Y: Integer);
begin
  MessageDlgPos(Msg, mtCustom, [mbOK], 0, X, Y);
end;

{ Input dialog }

function TTranslator.InputQuery(const ACaption, APrompt: string;
  var Value: string): Boolean;
var
  Form: TForm;
  Prompt: TLabel;
  Edit: TEdit;
  DialogUnits: TPoint;
  ButtonTop, ButtonWidth, ButtonHeight: Integer;
begin
  Result := False;
  Form := TForm.Create(Application);
  with Form do
  try
    Canvas.Font := Font;
    DialogUnits := GetAveCharSize(Canvas);
    BorderStyle := bsDialog;
    Caption := ACaption;
    ClientWidth := MulDiv(180, DialogUnits.X, 4);
    ClientHeight := MulDiv(63, DialogUnits.Y, 8);
    Position := poScreenCenter;
    Prompt := TLabel.Create(Form);
    with Prompt do
    begin
      Parent := Form;
      AutoSize := True;
      Left := MulDiv(8, DialogUnits.X, 4);
      Top := MulDiv(8, DialogUnits.Y, 8);
      Caption := APrompt;
    end;
    Edit := TEdit.Create(Form);
    with Edit do
    begin
      Parent := Form;
      Left := Prompt.Left;
      Top := MulDiv(19, DialogUnits.Y, 8);
      Width := MulDiv(164, DialogUnits.X, 4);
      MaxLength := 255;
      Text := Value;
      SelectAll;
    end;
    ButtonTop := MulDiv(41, DialogUnits.Y, 8);
    ButtonWidth := MulDiv(50, DialogUnits.X, 4);
    ButtonHeight := MulDiv(14, DialogUnits.Y, 8);
    with TButton.Create(Form) do
    begin
      Parent := Form;
      Caption := FButtonCaptions[mbOK] {SMsgDlgOK};
      ModalResult := mrOk;
      default := True;
      SetBounds(MulDiv(38, DialogUnits.X, 4), ButtonTop, ButtonWidth,
        ButtonHeight);
    end;
    with TButton.Create(Form) do
    begin
      Parent := Form;
      Caption := FButtonCaptions[mbCancel] {SMsgDlgCancel};
      ModalResult := mrCancel;
      Cancel := True;
      SetBounds(MulDiv(92, DialogUnits.X, 4), ButtonTop, ButtonWidth,
        ButtonHeight);
    end;
    if ShowModal = mrOk then
    begin
      Value := Edit.Text;
      Result := True;
    end;
  finally
    Form.Free;
  end;
end;

function TTranslator.InputBox(const ACaption, APrompt, ADefault: string): string;
begin
  Result := ADefault;
  InputQuery(ACaption, APrompt, Result);
end;

function _RemoveCRLF(const FCRChar, FLFChar, FTabChar: Char; S: string): string;
var
  I: Integer;
begin
  for I := 1 to Length(S) do
    if S[I] = #10 then
      S[I] := FLFChar
    else
      if S[I] = #13 then
        S[I] := FCRChar
      else
        if S[I] = #9 then S[I] := FTabChar;
  Result := S;
end;

function _AddCRLF(const FCRChar, FLFChar, FTabChar: Char; S: string): string;
var
  I: Integer;
begin
  for I := 1 to Length(S) do
    if S[I] = FLFChar then
      S[I] := #10
    else
      if S[I] = FCRChar then
        S[I] := #13
      else
        if S[I] = FTabChar then S[I] := #9;
  Result := S;
end;

function TranslateUserMessageFromIni(const FIniFile, FMessagesSection, sMsgName, sMessage: string; const FCRChar, FLFChar, FTabChar: Char): string; overload;
begin
  try
    with TMyIniFile.Create(FIniFile) do
    try
      Result := _AddCRLF(FCRChar, FLFChar, FTabChar, ReadString(FMessagesSection, _RemoveCRLF(FCRChar, FLFChar, FTabChar, sMsgName), '---- ---- ----'));
      if Result = '---- ---- ----' then
        Result := _AddCRLF(FCRChar, FLFChar, FTabChar, ReadString(FMessagesSection, _RemoveCRLF(FCRChar, FLFChar, FTabChar, sMessage), _RemoveCRLF(FCRChar, FLFChar, FTabChar, sMessage)));
    finally
      Free;
    end;
  except
    Result := sMessage;
  end;
end;

function TranslateUserMessageFromIni(const FIniFile, sMsgName, sMessage: string): string; overload;
begin
  try
    with TMyIniFile.Create(FIniFile) do
    try
      Result := _AddCRLF(#215, #247, #183, ReadString('Messages', _RemoveCRLF(#215, #247, #183, sMsgName), '---- ---- ----'));
      if Result = '---- ---- ----' then
        Result := _AddCRLF(#215, #247, #183, ReadString('Messages', _RemoveCRLF(#215, #247, #183, sMessage), _RemoveCRLF(#215, #247, #183, sMessage)));
    finally
      Free;
    end;
  except
    Result := sMessage;
  end;
end;

end.

